%{
/*
libfawk - A function-only AWK dialect (pascal compiler)

Copyright (c) 2019..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#define fpas_ctx_t fawk_ctx_t

#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <genht/hash.h>
#include "libfawk.h"
#include "parser.h"
#include "fpas.y.tab.h"
#include "parser_fpas.h"
#include "symtab.h"
#include "compile.h"
#include "execute.h"

#define FAWK_PUSH_IP() fawk_push_num(ctx, ctx->code.used)
#define FAWK_CURR_IP() ctx->code.used

#define pas_func_ret() \
	fawkc_addi(ctx, FAWKI_RET); \
	fawkc_addnum(ctx, ctx->compiler.numargs + (ctx->compiler.numfixedargs >= 0));

%}

%parse-param {fawk_ctx_t *ctx}
%lex-param {fawk_ctx_t *ctx}

%union {
	char *str;
	fawk_num_t num;
}

%token PT_NEQ PT_GTEQ PT_LTEQ
%token PT_NIL
%token PT_IF PT_THEN PT_ELSE
%token PT_DO PT_WHILE PT_REPEAT PT_UNTIL PT_FOR PT_DOWN PT_TO
%token PT_FUNCTION PT_BEGIN PT_END PT_CURRFUNC  PT_EXIT

%token <str> PT_ID PT_STRING
%token <num> PT_NUM

%type <num> for_to

%left PT_OR
%left PT_AND
%left PT_IN
%left '=' PT_NEQ PT_LTEQ '<' PT_GTEQ '>'
%left '@'
%left '+' '-'
%left '*' '/' '%'
%left PT_NOT
%right '(' ')'

%%

script:
	{ fawkc_addi(ctx, FAWKI_ABORT); fawkc_addcs(ctx, "uninitialized execute"); }
	program
	{ fawkc_addi(ctx, FAWKI_ABORT); fawkc_addcs(ctx, "ran beyond the script"); }
	;

program:
	func_def program
	| /* empty */
	;

/* statements to be used in normal context (statement in 'if' needs special casing) */
statement_noif:
	  s_assign
	| e_call          { fawkc_addi(ctx, FAWKI_POP); /* void experssion's result value shall be removed from the stack */ }
	| s_ret_assign
	| s_exit
	| s_loops
	| s_for
	| PT_BEGIN statements PT_END
	;

statement_in_if:
	  statement_noif
	| s_if
	;

statements:
	  /* empty */
	| statements statement
	;

/* A single statement, always terminated with ';' */
statement:
	  statement_noif ';'
	| s_if
	;

s_assign:
	var ':' '=' expr  { fawkc_addi(ctx, FAWKI_SET); fawkc_addi(ctx, FAWKI_PUSH_SYMVAL); fawkc_addi(ctx, FAWKI_POP); }
	;

/* The top of the stack between two statements is the function return value; pop it and allow expr to replace it */
s_ret_assign:
	PT_CURRFUNC ':' '=' { fawkc_addi(ctx, FAWKI_POP); }   expr
	;

s_exit:
	  PT_EXIT  { pas_func_ret(); }
	| PT_EXIT '('   { fawkc_addi(ctx, FAWKI_POP); }   expr ')'  { pas_func_ret(); }
	;


s_for:
	  s_for_plain
	| s_for_in
	;

s_for_in:
	PT_FOR var PT_IN var PT_DO
			{
				fawkc_addi(ctx, FAWKI_FORIN_FIRST);
				fawkc_addi(ctx, FAWKI_JMP); /* skip statements first, so next can load the variable first */
				fawkc_addnum(ctx, 0);
				FAWK_PUSH_IP();
			}
		statement_in_if
			{
				size_t begin = fawk_pop_num(ctx, 1);
				fawkc_addi(ctx, FAWKI_FORIN_NEXT);
				fawkc_addnum(ctx, begin);
				ctx->code.code[begin-1].data.num = FAWK_CURR_IP()-2;
			}
	;

s_for_plain:
	PT_FOR PT_ID ':' '='
			{
				fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
				fawkc_addsymref(ctx, $2, 0, ctx->compiler.funcdef_offs);
				fawkc_addnum(ctx, 0);
			}
			expr
			{
				fawkc_addi(ctx, FAWKI_SET);
				fawkc_addi(ctx, FAWKI_PUSH_SYMVAL);
				fawkc_addi(ctx, FAWKI_POP); /* void expr */

				fawkc_addi(ctx, FAWKI_JMP); /* jump-in: in the first iteration, skip condition and step and start at the body  */
				FAWK_PUSH_IP(); /* slot A */
				fawkc_addnum(ctx, 0);

				FAWK_PUSH_IP(); /* slot B */
			}
			for_to      /* hidden 3rd expr: stepping by +1 or -1 - shall be run at the end */
			{
		/* check range : if ((to-VAR)*step) >= 0, we can continue  - it's really  ((to-(VAR-step))*step) >= 0 because we step only after deciding */

				/* 'to' expr is already on the stack get the VAR and sub from 'to' */
				fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
				fawkc_addsymref(ctx, $2, 0, ctx->compiler.funcdef_offs);
				fawkc_addnum(ctx, 0);
				fawkc_addi(ctx, FAWKI_PUSH_SYMVAL);
				fawkc_addi(ctx, FAWKI_SUB);


				/* stepping is returned by 'for_to' */
				fawkc_addi(ctx, FAWKI_PUSH_NUM); fawkc_addnum(ctx, $8);

				/* need to dup the top 2 cells for the mul so step can be preserved for the actual stepping, plus swap them for adjusting VAR with step */
				fawkc_addi(ctx, FAWKI_PUSH_REL); fawkc_addnum(ctx, -1);
				fawkc_addi(ctx, FAWKI_PUSH_REL); fawkc_addnum(ctx, -3);

				/* step and (to-VAR) are on the stack */
				fawkc_addi(ctx, FAWKI_PUSH_REL); fawkc_addnum(ctx, -2);
				fawkc_addi(ctx, FAWKI_SUB);

				/* step and (to-(VAR-step)) are on the stack */
				fawkc_addi(ctx, FAWKI_MUL);

				/* push 0; stack is ((to-(VAR-step))*step),0 now */
				fawkc_addi(ctx, FAWKI_PUSH_NUM); fawkc_addnum(ctx, 0);

				fawkc_addi(ctx, FAWKI_GTEQ);
				fawkc_addi(ctx, FAWKI_POPJZ); /* jump out on false */
				FAWK_PUSH_IP(); /* slot C */

				fawkc_addnum(ctx, 0);

				/* step; stack is [1-] = step, [-2] = ((to-(VAR-step))*step) */

				fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
				fawkc_addsymref(ctx, $2, 0, ctx->compiler.funcdef_offs);
				fawkc_addnum(ctx, 0);
				fawkc_addi(ctx, FAWKI_PUSH_TOPVAR);
				fawkc_addi(ctx, FAWKI_PUSH_REL); fawkc_addnum(ctx, -3);  /* [-1]=VAR, [-2]=step, [-3]=... */
				fawkc_addi(ctx, FAWKI_ADD); fawkc_addi(ctx, FAWKI_SET);

				fawkc_addi(ctx, FAWKI_POP); /* void expr (stepping) */
				fawkc_addi(ctx, FAWKI_POP); /* remove step */
				fawkc_addi(ctx, FAWKI_POP); /* remove the original ((to-VAR)*step)*/

				fawkc_addi(ctx, FAWKI_JMP); /* skip 3rd expr */
				FAWK_PUSH_IP(); /* slot D */
				fawkc_addnum(ctx, 0);


				fawkc_addi(ctx, FAWKI_JMP); /* jump back to the check */
				FAWK_PUSH_IP(); /* slot E */
				fawkc_addnum(ctx, 0);
			}
			PT_DO statement_in_if
			{
				size_t jumpin, jback3rd, jskip, jout, rerun;

				jback3rd = fawk_pop_num(ctx, 1); /* slot E */
				jskip = fawk_pop_num(ctx, 1);    /* slot D */
				jout = fawk_pop_num(ctx, 1);     /* slot C */
				rerun = fawk_pop_num(ctx, 1);    /* slot B */
				jumpin = fawk_pop_num(ctx, 1);   /* slot A */

				fawkc_addi(ctx, FAWKI_JMP); /* jump back to execute the 3rd expr */
				fawkc_addnum(ctx, jskip+1);

				ctx->code.code[jumpin].data.num = jback3rd+1;
				ctx->code.code[jback3rd].data.num = rerun;
				ctx->code.code[jskip].data.num = jback3rd+1;
				ctx->code.code[jout].data.num = FAWK_CURR_IP();
				fawkc_addi(ctx, FAWKI_POP);
				fawkc_addi(ctx, FAWKI_POP);
			}
	;

/* push stepping at the end */
for_to:
	  PT_TO expr { $$ = +1; }
	| PT_DOWN PT_TO expr { $$ = -1; }
	;

s_loops:
	  s_while_do
	| s_repeat_until
	;

s_while_do:
	PT_WHILE                         { FAWK_PUSH_IP(); }
		expr PT_DO                     { fawkc_addi(ctx, FAWKI_POPJZ); FAWK_PUSH_IP(); fawkc_addnum(ctx, 0); }
		statement_in_if                { loop_pretest_jumpback(); }
	;

s_repeat_until:
	PT_REPEAT                        { FAWK_PUSH_IP(); /*jump back here to repeat */ }
		statements
		PT_UNTIL expr                  { fawkc_addi(ctx, FAWKI_POPJZ); fawkc_addnum(ctx, fawk_pop_num(ctx, 1));}
		;

s_if:
		if_then statement_in_if ';'
			{
				size_t jmp1 = fawk_pop_num(ctx, 1);
				ctx->code.code[jmp1].data.num = FAWK_CURR_IP();
			}
		| if_then statement_in_if PT_ELSE
		{
			fawkc_addi(ctx, FAWKI_JMP);
			FAWK_PUSH_IP(); fawkc_addnum(ctx, 777);
		}
		statement
		{
			size_t jmp_then_post = fawk_pop_num(ctx, 1), jmp_if = fawk_pop_num(ctx, 1);
			ctx->code.code[jmp_then_post].data.num = FAWK_CURR_IP();
			ctx->code.code[jmp_if].data.num = jmp_then_post+1;
		}
	;

if_then:
	PT_IF expr PT_THEN    { fawkc_addi(ctx, FAWKI_POPJZ); FAWK_PUSH_IP(); fawkc_addnum(ctx, 777); }
	;

func_def:
	PT_FUNCTION PT_ID                  { ctx->compiler.numfixedargs = -1; ctx->compiler.numargs = 0; }
		'(' arg_def ')' ';'              { fawk_symtab_regfunc(ctx, $2, FAWK_CURR_IP(), ctx->compiler.numargs, ctx->compiler.numfixedargs); ctx->fp = ctx->sp; ctx->parser.curr_func = $2; fawkc_addi(ctx, FAWKI_PUSH_NIL); /* this is going to be the return value */ }
		PT_BEGIN statements PT_END ';'
		{
			int n;
			pas_func_ret();
			for(n = 0; n < ctx->compiler.numargs + (ctx->compiler.numfixedargs >= 0); n++) {
				fawk_cell_t cell;
				fawk_pop(ctx, &cell);
				fawk_cell_free(ctx, &cell);
			}
			ctx->fp = 0;
			free(ctx->parser.curr_func); ctx->parser.curr_func = NULL;
		}
	;

arg_def:
	  /* empty */
	| PT_ID                                      { fawk_push_str(ctx, $1); fawk_free(ctx, $1); ctx->compiler.numargs++; }
	| '.' '.' '.'                                { ctx->compiler.numfixedargs = ctx->compiler.numargs; fawk_push_str(ctx, "VARARG"); }
	| arg_def ';' PT_ID                          { fawk_push_str(ctx, $3); fawk_free(ctx, $3); ctx->compiler.numargs++; }
	| arg_def ';' '.' '.' '.'                    { ctx->compiler.numfixedargs = ctx->compiler.numargs; fawk_push_str(ctx, "VARARG"); }
	;


expr:
	  PT_NUM                       { fawkc_addi(ctx, FAWKI_PUSH_NUM); fawkc_addnum(ctx, $1); }
	| PT_STRING                    { fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $1); fawk_free(ctx, $1); }
	| PT_NIL                       { fawkc_addi(ctx, FAWKI_PUSH_NIL); }
	| var                          { fawkc_addi(ctx, FAWKI_PUSH_SYMVAL); }
	| '&' var                      /* done in var */
	| e_call
	| '(' expr ')'                /* done implicitly by the parser */
	| expr '=' expr               { fawkc_addi(ctx, FAWKI_EQ); }
	| expr PT_NEQ expr            { fawkc_addi(ctx, FAWKI_NEQ); }
	| expr PT_GTEQ expr           { fawkc_addi(ctx, FAWKI_GTEQ); }
	| expr PT_LTEQ expr           { fawkc_addi(ctx, FAWKI_LTEQ); }
	| expr '>' expr               { fawkc_addi(ctx, FAWKI_GT); }
	| expr '<' expr               { fawkc_addi(ctx, FAWKI_LT); }
	| expr PT_IN var              { fawkc_addi(ctx, FAWKI_IN); }
	| e_or
	| e_and
	| expr '+' expr               { fawkc_addi(ctx, FAWKI_ADD); }
	| expr '-' expr               { fawkc_addi(ctx, FAWKI_SUB); }
	| expr '*' expr               { fawkc_addi(ctx, FAWKI_MUL); }
	| expr '/' expr               { fawkc_addi(ctx, FAWKI_DIV); }
	| expr '%' expr               { fawkc_addi(ctx, FAWKI_MOD); }
	| '-' expr                    { fawkc_addi(ctx, FAWKI_NEG); }
	| '+' expr                    /* a positive number is the same number, don't do anything */
	| PT_NOT expr                 { fawkc_addi(ctx, FAWKI_NOT); }
	| expr '@' expr               { fawkc_addi(ctx, FAWKI_CONCAT); }
	;

var:
	PT_ID          { fawk_push_num(ctx, ctx->compiler.numidx); ctx->compiler.numidx = 0; }
		arridx
		{
			fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
			fawkc_addsymref(ctx, $1, ctx->compiler.numidx, 0);
			fawkc_addnum(ctx, ctx->compiler.numidx);
			fawk_free(ctx, $1);
			ctx->compiler.numidx = fawk_pop_num(ctx, 1);
		}
	;

arridx:
	  /* empty */
	| arridx '[' ']'           { ctx->compiler.numidx = -1; }
	| arridx '[' aix_expr ']'  { ctx->compiler.numidx++; }
	| arridx '.' PT_ID         { ctx->compiler.numidx++; fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $3); fawk_free(ctx, $3); }
	| arridx '.' PT_STRING     { ctx->compiler.numidx++; fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $3); fawk_free(ctx, $3); }
	;

aix_expr: /* array index expression */
	  expr
	| aix_expr ','   { parse_aix_expr(); }
		expr           { fawkc_addi(ctx, FAWKI_CONCAT); }
	;

/* lazy OR evaluation */
e_or:
	expr PT_OR             { lazy_binop1(ctx, 1); }
		expr                 { lazy_binop2(ctx, 1); }
	;

/* lazy AND evaluation */
e_and:
	expr PT_AND             { lazy_binop1(ctx, 0); }
		expr                  { lazy_binop2(ctx, 0); }
	;

e_call:
	expr '('                      { fawk_push_num(ctx, ctx->compiler.numargs); ctx->compiler.numargs = 0; ctx->code.used--; /* remove push_symval on the func symref */ }
		call_arg ')'
		{
			size_t old_numargs;

			old_numargs = fawk_pop_num(ctx, 1);

			fawkc_addi(ctx, FAWKI_CALL);
			fawkc_addnum(ctx, ctx->compiler.numargs);
			ctx->compiler.numargs = old_numargs;
		}
	;

call_arg:
	  /* empty */
	| expr                 { ctx->compiler.numargs++; }
	| expr ',' call_arg    { ctx->compiler.numargs++; }
	;

%%

FAWK_API void fpas_error(fpas_ctx_t *ctx, fpas_STYPE tok, const char *msg) { libfawk_error(ctx, msg, ctx->parser.isp->fn, ctx->parser.isp->line+1, ctx->parser.isp->col+1); }

FAWK_API int fawk_lex_fpas(fpas_STYPE *lval, fawk_ctx_t *ctx);
FAWK_API int fawk_parse_fpas(fawk_ctx_t *ctx) { fawk_parser_loop(fpas_yyctx_t, fpas_STYPE, fawk_lex_fpas, fpas_parse, ctx, fpas_RES_NEXT, fpas_RES_DONE); }
